/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
/*==========================================================================
 * 
 * Project : Freyja
 * Author  : Terry 'Mongoose' Hendrix II
 * Website : http://gooseegg.sourceforge.net
 * Email   : mongoose@icculus.org
 * Object  : 
 * Comments: This is the mel plug-in module
 *
 *           See file COPYING for license details.
 *
 *
 *-- Test Defines -----------------------------------------------------
 *           
 *
 *-- History ---------------------------------------------------------- 
 *
 * 2006.08.15:
 * Mongoose - Created
 ==========================================================================*/

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <freyja/ModelABI.h>
#include <freyja/PluginABI.h>
#include <freyja/BoneABI.h>
#include <freyja/SkeletonABI.h>
#include <freyja/TextureABI.h>
#include <freyja/LegacyABI.h>
#include <freyja/MeshABI.h>
#include <freyja/freyja.h>
#include <mstl/Vector.h>
#include <mstl/SystemIO.h>

using namespace mstl;

extern "C" {

	int freyja_model__ma_check(char *filename);
	int freyja_model__ma_import(char *filename);
	int freyja_model__ma_export(char *filename);
	int import_model(char *filename);
	void freyja_init();
}

void freyja_init()
{
	freyjaPluginName1s("ma");
	freyjaPluginDescription1s("Maya ASCII Scene (*.ma)");
	freyjaPluginAddExtention1s("*.ma");
	freyjaPluginImport1i(FREYJA_PLUGIN_SKELETON | FREYJA_PLUGIN_MESH);
	freyjaPluginExport1i(FREYJA_PLUGIN_SKELETON | FREYJA_PLUGIN_MESH);
}

int import_model(char *filename)
{
	if (!freyja_model__ma_check(filename))
		return freyja_model__ma_import(filename);

	return -1;
}


int freyja_model__ma_check(char *filename)
{
	if (SystemIO::File::CompareFilenameExtention(filename, ".ma") == 0 ||
		SystemIO::File::CompareFilenameExtention(filename, ".MA") == 0)
	{
		freyjaPrintMessage("mel.so: '%s' is an MA file.", filename);
		return 0;
	}

	return -1;
}


int freyja_model__ma_import(char *filename)
{
	SystemIO::TextFileReader r;
	Vector <unsigned int> transV;
	Vector <unsigned int> transT;
	Vector <vec_t> normals;
	//Vector <unsigned int> transN;
	const char *symbol;
	vec_t x, y, z, u, v;
	int i, vertexIndex, index;
	bool hasUV = false;
	bool hasNormal = false;


	if (!r.Open(filename))
	{
		return -1;
	}

	freyjaBegin(FREYJA_MODEL);

	while ((symbol = r.ParseSymbol()) && !r.IsEndOfFile())
	{
		// Comment
		if (!strncmp(symbol, "//", 2))
		{
			char c;
			while ((c = r.ReadChar()) && !r.IsEndOfFile())
			{
				if (c == '\n')
					break;
			}
		}
		// requires maya "3.0";
		else if (!strcmp(symbol, "requires"))
		{
			r.FindNextChar(';');
		}
		// currentUnit -l centimeter -a degree -t film;
		else if (!strcmp(symbol, "currentUnit"))
		{
			r.FindNextChar(';');
		}
		else if (!strcmp(symbol, "createNode"))
		{
			// createNode transform -s -n "persp";
			if (!strcmp(symbol, "transform"))
			{
				r.FindNextChar(';');
			}
			// createNode mesh -n "r_meshShape" -p "r_mesh";
			else if (!strcmp(symbol, "mesh"))
			{
				freyjaBegin(FREYJA_MESH);
				r.FindNextChar(';');
				long offset = r.GetOffset();

				while ((symbol = r.ParseSymbol()) && !r.IsEndOfFile())
				{
					if (!strcmp(symbol, "setAttr"))
					{
						symbol = r.ParseSymbol();

						if (!strcmp(symbol, "-s"))
						{
							long count = r.ParseInteger();
							const char *s = r.ParseStringLiteral();

							if (s)
							{
								delete [] s;
							}
						}

						r.FindNextChar(';');
					}
					else
					{
						// Rewind if not "setAttr"
						r.SetOffset(offset);
						break;
					}

					offset = r.GetOffset();
				}

				freyjaEnd(); // FREYJA_MESH
			}
		}
		else if (!strcmp(symbol, "vt"))
		{
			u = r.ParseFloat();
			v = r.ParseFloat();
			hasUV = true;

			transT.pushBack(freyjaTexCoordCreate2f(u, v));
		}
		else if (!strcmp(symbol, "vn"))  // Assumes normals come after v
		{
			x = r.ParseFloat();
			y = r.ParseFloat();
			z = r.ParseFloat();
			hasNormal = true;

			// polymapped normals not in Freyja
			normals.pushBack(x);
			normals.pushBack(y);
			normals.pushBack(z);
		}
		else if (!strcmp(symbol, "v"))
		{
			x = r.ParseFloat();
			y = r.ParseFloat();
			z = r.ParseFloat();

			transV.pushBack(freyjaVertexCreate3f(x, y, z));
		}
		else if (!strcmp(symbol, "o"))
		{
			// Name of the new group, but no used here
			symbol = r.ParseSymbol();
			freyjaPrintMessage("FIXME: Object name '%s' is not used\n", symbol);
		}
		else if (!strcmp(symbol, "g"))
		{
			// Name of the new group, but no used here
			symbol = r.ParseSymbol();
			freyjaPrintMessage("FIXME: Group name '%s' is not used\n", symbol);
		}
		else if (!strcmp(symbol, "s"))
		{
			index = r.ParseInteger();
			freyjaPrintMessage("FIXME: Smoothing Group %i is not used\n", i);
		}
		else if (!strcmp(symbol, "f")) // f 1/1/1 2/2/2 3/3/3
		{
			// Start a new polygon
			freyjaBegin(FREYJA_POLYGON);

			// FIXME: Add state machine parser to handle all the quirks like 1//1

			for (i = 0; i < 3; ++i)
			{
				index = r.ParseInteger();
				freyjaPolygonVertex1i(transV[index-1]); // index starts at 1 not 0
				vertexIndex = transV[index-1];

				if (hasUV)
				{
					r.ReadChar(); // '/' no whitespace
					index = r.ParseInteger();

					index = transT[index-1]; // index starts at 1 not 0
					freyjaPolygonTexCoord1i(index);
				}
				else
				{
					freyjaVertexTexcoord2f(index, 0.25*i, 0.25*(i%2));
				}

				if (hasNormal) // polymapped normals aren't in Freyja atm
				{
					r.ReadChar(); // '/' no whitespace
					index = r.ParseInteger();

					//index = transN[index-1]; // index starts at 1 not 0
					//freyjaNormal1i(index);

					if (!normals.empty())
					{
						if ((index-1)*3+2 > (int)normals.size())
						{
							freyjaPrintError("obj.so> bad normal parse?");
						}
						else
						{
							freyjaVertexNormal3f(vertexIndex,
												 normals[(index-1)*3],
												 normals[(index-1)*3+1],
												 normals[(index-1)*3+2]);
						}
					}
				}
			}

			// Handle quad faces
			i = r.GetOffset();
			symbol = r.ParseSymbol();
			r.SetOffset(i);

			if (isdigit(symbol[0]))
			{
				index = r.ParseInteger();
				freyjaPolygonVertex1i(transV[index-1]); // index starts at 1 not 0

				if (hasUV)
				{
					r.ReadChar(); // '/' no whitespace
					index = r.ParseInteger();

					index = transT[index-1]; // index starts at 1 not 0
					freyjaPolygonTexCoord1i(index);
				}
				else
				{
					freyjaVertexTexcoord2f(index, 0.25*i, 0.25*(i%2));
				}

				if (hasNormal) // polymapped normals aren't in Freyja atm
				{
					r.ReadChar(); // '/' no whitespace
					index = r.ParseInteger();

					//index = transN[index-1]; // index starts at 1 not 0
					//freyjaNormal1i(index);
				}
			}

			freyjaPolygonMaterial1i(0);

			freyjaEnd(); // FREYJA_POLYGON
		}
	}

	freyjaEnd(); // FREYJA_MODEL

	return 0;
}


#if 0
bool saveMEL (const char *filename)
{
	FILE *file = fopen (filename, "wt");
	if (!file)
		return false;

	int i;
	// writing joints
	if (g_numFrames == 1)
	{
		for (i = 0; i < g_numBones; i++)
		{
			if (g_bones[i].parentId != -1)
				fprintf (file, "select -r %s ;\n", g_bones[g_bones[i].parentId].name);

			fprintf (file, "joint -a -n %s ;\n", g_bones[i].name);
		}
	}

	//fprintf (file, "playbackOptions -min 0 -max %d -ast 1 -aet %d -by 1 ;\n", g_numFrames, g_numFrames);

	// writing key frames
	if (g_numFrames > 1)
		g_numFrames--;
	for (int j = 0; j < g_numFrames; j++)
	{
		for (int i = 0; i < g_numBones; i++)
		{
			if (g_bones[i].parentId != -1)
			{
				fprintf (file, "setKeyframe -v %f -t %f -at translateX %s ;\n", g_bones[i].pos[j][0], (float) j * g_fScale + (float) g_nStartFrame, g_bones[i].name);
				fprintf (file, "setKeyframe -v %f -t %f -at translateY %s ;\n", g_bones[i].pos[j][1], (float) j * g_fScale + (float) g_nStartFrame, g_bones[i].name);
				fprintf (file, "setKeyframe -v %f -t %f -at translateZ %s ;\n", g_bones[i].pos[j][2], (float) j * g_fScale + (float) g_nStartFrame, g_bones[i].name);
				fprintf (file, "setKeyframe -v %f -t %f -at rotateX %s ;\n", g_bones[i].rot[j][0] * (180.0f / (float) Q_PI), (float) j * g_fScale + (float) g_nStartFrame, g_bones[i].name);
				fprintf (file, "setKeyframe -v %f -t %f -at rotateY %s ;\n", g_bones[i].rot[j][1] * (180.0f / (float) Q_PI), (float) j * g_fScale + (float) g_nStartFrame, g_bones[i].name);
				fprintf (file, "setKeyframe -v %f -t %f -at rotateZ %s ;\n", g_bones[i].rot[j][2] * (180.0f / (float) Q_PI), (float) j * g_fScale + (float) g_nStartFrame, g_bones[i].name);
			}
		}
	}

	fclose (file);

	return true;
}
#endif


int freyja_model__ma_export(char *filename)
{
	const vec_t scale = 1.0;
	SystemIO::TextFileWriter w;
	long modelIndex = 0; // make plugin option
	long i, j, k, v;
	long meshCount, meshIndex;
	long vertexCount, vertexIndex;
	long polygonCount, polygonIndex, faceVertexCount;
	vec3_t xyz;
	vec2_t uv;


	if (!w.Open(filename))
	{
		return -1;
	}

	/* Header comments */
	w.Print("//Maya ASCII 1.0 scene\n");
	w.Print("//Name: %s\n", filename);
	w.Print("//Last modified: %s, %s %i, %i %i:%%i:%i AM\n",
			"Mon", "Jan", 1, 99, 0, 0, 0);


	/* External files */
	//w.Print("file -r -rpr \"freyja\" \"/freyja\";\n");

	/* Requirements */
	w.Print("requires maya \"2.0\";\n");

	/* Units */
	w.Print("currentUnit -l cm -a radians -t ntsc;\n");

	/* Meshes */
	meshCount = freyjaGetModelMeshCount(modelIndex);
	v = 0;

	for (i = 0; i < meshCount; ++i)
	{
		meshIndex = freyjaGetModelMeshIndex(modelIndex, i);

		vertexCount = freyjaGetMeshVertexCount(meshIndex);
		polygonCount = freyjaGetMeshPolygonCount(meshIndex);

		w.Print("createNode mesh -n \"mesh%iShape\" -p \"mesh%i\";\n", i);

		/* Vertices */
		w.Print("setAttr -s %i \".vt[0:%i]\"\n", vertexCount, vertexCount-1);
		for (j = 0; j < vertexCount; ++j)
		{
			vertexIndex = j;//freyjaGetMeshVertexIndex(meshIndex, j);
			freyjaGetVertexXYZ3fv(vertexIndex, xyz);

			//w.Print("# vertex %li -> %li\n", j, vertexIndex);
			w.Print("%f %f %f\n",
					xyz[0]*scale, xyz[1]*scale, xyz[2]*scale);
		}

		/* Normals */
		w.Print("\n# normalCount %li\n", vertexCount);
		for (j = 0; j < vertexCount; ++j)
		{
			vertexIndex = j;//freyjaGetMeshVertexIndex(meshIndex, j);
			freyjaGetVertexNormalXYZ3fv(vertexIndex, xyz);

			w.Print("vn %f %f %f\n",
					xyz[0], xyz[1], xyz[2]);
		}

		/* Texcoords */
		w.Print("setAttr -s %i \".uvst[0].uvsp[0:%i]\" -type \"float2\"\n",
				vertexCount, vertexCount-1);
		for (j = 0; j < vertexCount; ++j)
		{
			vertexIndex = j;//freyjaGetMeshVertexIndex(meshIndex, j);
			freyjaGetVertexTexcoord2fv(vertexIndex, uv);

			w.Print("%f %f\n", uv[0], uv[1]);
		}

		/* Edges*/
#error FIXME Maya ascii uses faces composed of edges not vertices
		w.Print("setAttr -s %i \".ed[0:%i]\" -type \"polyFaces\" \n", 
				edgeCount, edgeCount-1);

		/* Faces */
		w.Print("setAttr -s %i \".fc[0:%i]\" -type \"polyFaces\" \n", 
				polygonCount, polygonCount-1);
		for (j = 0; j < polygonCount; ++j)
		{
			polygonIndex = j;//freyjaGetMeshPolygonIndex(meshIndex, j);
			faceVertexCount = freyjaGetPolygonVertexCount(polygonIndex);

			w.Print("f ");

			for (k = 0; k < faceVertexCount; ++k)
			{
				//faceVertex = freyjaGetPolygonVertexIndex(polygonIndex, k);
				vertexIndex = freyjaGetMeshPolygonVertexIndex(meshIndex, j, k);

				//w.Print("\n# %li. %li -> %li\n", j, faceVertex, vertexIndex);

				// Assumes UV mapped vertices exported
				// Remember obj indices start at 1
				w.Print("%li/%li/%li ", 
						v+vertexIndex+1, v+vertexIndex+1, v+vertexIndex+1);
			}

			w.Print("\n");			
		}

		v += vertexCount;
	}

	w.Close();

	return 0;
}

